//cpu8040_base.h
//CPU emulation base class

#ifndef __XATREMU_CPU8040_BASE__
#define __XATREMU_CPU8040_BASE__

#include <stdint.h>
#include "testdefines.h"

namespace xatremu
{

class WD1772Base;
class SerialBase;

#define MAXCYCLE 0x7fffffffffffffff
class CPU8040Base
{
public:
	CPU8040Base();
	virtual ~CPU8040Base();
	virtual void CheckCPUSpecials();
	//disassemble next instruction in program memory
	virtual void ExecuteInstruction();
	virtual void SetNextSpecialEvent(unsigned char eventType, int64_t addCycles);
	virtual void RunCPU();
	virtual void RunCPUForCycles(unsigned int numOfCycles);

#define SPECIALEVENTCOUNT 8
#define CPUEVENTTIMERINCREASE 0
#define CPUEVENTSERIALIN 1
#define CPUEVENTSERIALOUT 2
#define CPUEVENTWD1772MOTOR 3
#define CPUEVENTWD1772HEADSTEP 4
#define CPUEVENTWD1772HEADCALM 5
#define CPUEVENTWD1772NEXTDATA 6
#define CPUEVENTWD1772DATATIMEOUT 7

protected:
	virtual void CheckCPUSpecialEvents();
	virtual void DetermineNextEventCount(void);
	//table with instruction emulation subroutines
	//typedef void (CPU8040::* CPUFunc)(void);
	//CPUFunc CPUFuncTable[256];

protected:
	//unsigned short r_PC;
	//unsigned char r_R0, r_R1, r_R2, r_R3, r_R4, r_R5, r_R6, r_R7;
	int64_t m_totalCycles, m_storedCycles; //since reset, takes 30000 years to overflow
	//unsigned char r_A;
	//unsigned char r_psw;
	//unsigned char r_sp;
	//bool r_dbf;
	//bool r_C;
	//r_AC ignored, XF551 rom doesnt use instruction DA A
	//bool r_F0, r_F1;
	//bool r_TF; //timer flag
	//unsigned char r_T; //timer value
	//bool m_TRunning; //timer running?
	//bool m_ExternalIntEnabled, m_TCNTIntEnabled, m_IntPossible;
	//void *m_instrMemPtr;
	//unsigned long m_ROMLen;
	WD1772Base* m_WD1772;
	SerialBase* m_Serial;
	//unsigned char m_internalRAM[0x100];
	//unsigned char m_internalStack[0x100]; //maybe incorrectly separated, lets see
	int64_t m_NextSpecialEvent;
	//events: Timer; Serial In; Serial Out
	int64_t m_SpecialEventTable[SPECIALEVENTCOUNT];

	/*
	Port 1:
	Bit 0, 1: Register select WD1772 (out)
	Bit 2: INTR WD1772 (in)
	Bit 3: 'DD WD1772 (out)
	Bit 4: 'MR WD1772 (out)
	Bit 5: R/'W WD1772 (out)
	Bit 6: Side Floppy (out)
	Bit 7: Serial Out (out)
	Port 2:
	Bit 0-5: Not connected, maintain their state
	Bit 6, 7: Floppy station number (in)
	*/
	unsigned char r_Port1, r_Port2;
	unsigned char m_DriveNumber;	//0-3

	/*
	T0: Serial data incoming from Atari computer (in)
	T1: DRQ WD1772 (in)
	INT: Command line SIO (in)
	*/

	virtual void I_ADD_A_AT_R0(void);
	virtual void I_ADD_A_AT_R1(void);
	virtual void I_ADD_A_IMM(void);
	virtual void I_ADD_A_R0(void);
	virtual void I_ADD_A_R1(void);
	virtual void I_ADD_A_R2(void);
	virtual void I_ADD_A_R3(void);
	virtual void I_ADD_A_R4(void);
	virtual void I_ADD_A_R5(void);
	virtual void I_ADD_A_R6(void);
	virtual void I_ADD_A_R7(void);
	virtual void I_ADDC_A_AT_R0(void);
	virtual void I_ADDC_A_AT_R1(void);
	virtual void I_ADDC_A_IMM(void);
	virtual void I_ADDC_A_R0(void);
	virtual void I_ADDC_A_R1(void);
	virtual void I_ADDC_A_R2(void);
	virtual void I_ADDC_A_R3(void);
	virtual void I_ADDC_A_R4(void);
	virtual void I_ADDC_A_R5(void);
	virtual void I_ADDC_A_R6(void);
	virtual void I_ADDC_A_R7(void);
	virtual void I_ANL_A_AT_R0(void);
	virtual void I_ANL_A_AT_R1(void);
	virtual void I_ANL_A_IMM(void);
	virtual void I_ANL_A_R0(void);
	virtual void I_ANL_A_R1(void);
	virtual void I_ANL_A_R2(void);
	virtual void I_ANL_A_R3(void);
	virtual void I_ANL_A_R4(void);
	virtual void I_ANL_A_R5(void);
	virtual void I_ANL_A_R6(void);
	virtual void I_ANL_A_R7(void);
	virtual void I_ANL_BUS_IMM(void);
	virtual void I_ANL_P1_IMM(void);
	virtual void I_ANL_P2_IMM(void);
	virtual void I_ANLD_P4_A(void);
	virtual void I_ANLD_P5_A(void);
	virtual void I_ANLD_P6_A(void);
	virtual void I_ANLD_P7_A(void);
	virtual void I_CALL_CODE(void);
	virtual void I_CLR_A(void);
	virtual void I_CLR_C(void);
	virtual void I_CLR_F0(void);
	virtual void I_CLR_F1(void);
	virtual void I_CPL_A(void);
	virtual void I_CPL_C(void);
	virtual void I_CPL_F0(void);
	virtual void I_CPL_F1(void);
	virtual void I_DA_A(void);
	virtual void I_DEC_A(void);
	virtual void I_DEC_R0(void);
	virtual void I_DEC_R1(void);
	virtual void I_DEC_R2(void);
	virtual void I_DEC_R3(void);
	virtual void I_DEC_R4(void);
	virtual void I_DEC_R5(void);
	virtual void I_DEC_R6(void);
	virtual void I_DEC_R7(void);
	virtual void I_DIS_I(void);
	virtual void I_DIS_TCNTI(void);
	virtual void I_DJNZ_R0_CODE(void);
	virtual void I_DJNZ_R1_CODE(void);
	virtual void I_DJNZ_R2_CODE(void);
	virtual void I_DJNZ_R3_CODE(void);
	virtual void I_DJNZ_R4_CODE(void);
	virtual void I_DJNZ_R5_CODE(void);
	virtual void I_DJNZ_R6_CODE(void);
	virtual void I_DJNZ_R7_CODE(void);
	virtual void I_EN_I(void);
	virtual void I_EN_TCNTI(void);
	virtual void I_ENT0_CLK(void);
	virtual void I_IN_A_P1(void);
	virtual void I_IN_A_P2(void);
	virtual void I_INC_A(void);
	virtual void I_INC_AT_R0(void);
	virtual void I_INC_AT_R1(void);
	virtual void I_INC_R0(void);
	virtual void I_INC_R1(void);
	virtual void I_INC_R2(void);
	virtual void I_INC_R3(void);
	virtual void I_INC_R4(void);
	virtual void I_INC_R5(void);
	virtual void I_INC_R6(void);
	virtual void I_INC_R7(void);
	virtual void I_INS_A_BUS(void);
	virtual void I_JB0_CODE(void);
	virtual void I_JB1_CODE(void);
	virtual void I_JB2_CODE(void);
	virtual void I_JB3_CODE(void);
	virtual void I_JB4_CODE(void);
	virtual void I_JB5_CODE(void);
	virtual void I_JB6_CODE(void);
	virtual void I_JB7_CODE(void);
	virtual void I_JC_CODE(void);
	virtual void I_JF0_CODE(void);
	virtual void I_JF1_CODE(void);
	virtual void I_JMP_CODE(void);
	virtual void I_JMPP_AT_A(void);
	virtual void I_JNC_CODE(void);
	virtual void I_JNI_CODE(void);
	virtual void I_JNT0_CODE(void);
	virtual void I_JNT1_CODE(void);
	virtual void I_JNZ_CODE(void);
	virtual void I_JT0_CODE(void);
	virtual void I_JT1_CODE(void);
	virtual void I_JTF_CODE(void);
	virtual void I_JZ_CODE(void);
	virtual void I_MOV_A_IMM(void);
	virtual void I_MOV_A_AT_R0(void);
	virtual void I_MOV_A_AT_R1(void);
	virtual void I_MOV_A_PSW(void);
	virtual void I_MOV_A_R0(void);
	virtual void I_MOV_A_R1(void);
	virtual void I_MOV_A_R2(void);
	virtual void I_MOV_A_R3(void);
	virtual void I_MOV_A_R4(void);
	virtual void I_MOV_A_R5(void);
	virtual void I_MOV_A_R6(void);
	virtual void I_MOV_A_R7(void);
	virtual void I_MOV_A_T(void);
	virtual void I_MOV_AT_R0_A(void);
	virtual void I_MOV_AT_R1_A(void);
	virtual void I_MOV_AT_R0_IMM(void);
	virtual void I_MOV_AT_R1_IMM(void);
	virtual void I_MOV_PSW_A(void);
	virtual void I_MOV_R0_A(void);
	virtual void I_MOV_R1_A(void);
	virtual void I_MOV_R2_A(void);
	virtual void I_MOV_R3_A(void);
	virtual void I_MOV_R4_A(void);
	virtual void I_MOV_R5_A(void);
	virtual void I_MOV_R6_A(void);
	virtual void I_MOV_R7_A(void);
	virtual void I_MOV_R0_IMM(void);
	virtual void I_MOV_R1_IMM(void);
	virtual void I_MOV_R2_IMM(void);
	virtual void I_MOV_R3_IMM(void);
	virtual void I_MOV_R4_IMM(void);
	virtual void I_MOV_R5_IMM(void);
	virtual void I_MOV_R6_IMM(void);
	virtual void I_MOV_R7_IMM(void);
	virtual void I_MOV_T_A(void);
	virtual void I_MOVD_A_P4(void);
	virtual void I_MOVD_A_P5(void);
	virtual void I_MOVD_A_P6(void);
	virtual void I_MOVD_A_P7(void);
	virtual void I_MOVD_P4_A(void);
	virtual void I_MOVD_P5_A(void);
	virtual void I_MOVD_P6_A(void);
	virtual void I_MOVD_P7_A(void);
	virtual void I_MOVP_A_AT_A(void);
	virtual void I_MOVP3_A_AT_A(void);
	virtual void I_MOVX_A_AT_R0(void);
	virtual void I_MOVX_A_AT_R1(void);
	virtual void I_MOVX_AT_R0_A(void);
	virtual void I_MOVX_AT_R1_A(void);
	virtual void I_NOP(void);
	virtual void I_ORL_A_AT_R0(void);
	virtual void I_ORL_A_AT_R1(void);
	virtual void I_ORL_A_IMM(void);
	virtual void I_ORL_A_R0(void);
	virtual void I_ORL_A_R1(void);
	virtual void I_ORL_A_R2(void);
	virtual void I_ORL_A_R3(void);
	virtual void I_ORL_A_R4(void);
	virtual void I_ORL_A_R5(void);
	virtual void I_ORL_A_R6(void);
	virtual void I_ORL_A_R7(void);
	virtual void I_ORL_BUS_IMM(void);
	virtual void I_ORL_P1_IMM(void);
	virtual void I_ORL_P2_IMM(void);
	virtual void I_ORLD_P4_A(void);
	virtual void I_ORLD_P5_A(void);
	virtual void I_ORLD_P6_A(void);
	virtual void I_ORLD_P7_A(void);
	virtual void I_OUTL_BUS_A(void);
	virtual void I_OUTL_P1_A(void);
	virtual void I_OUTL_P2_A(void);
	virtual void I_RET(void);
	virtual void I_RETR(void);
	virtual void I_RL_A(void);
	virtual void I_RLC_A(void);
	virtual void I_RR_A(void);
	virtual void I_RRC_A(void);
	virtual void I_SEL_MB0(void);
	virtual void I_SEL_MB1(void);
	virtual void I_SEL_RB0(void);
	virtual void I_SEL_RB1(void);
	virtual void I_STOP_TCNT(void);
	virtual void I_STRT_CNT(void);
	virtual void I_STRT_T(void);
	virtual void I_SWAP_A(void);
	virtual void I_UNDEF(void); //new
	virtual void I_XCH_A_AT_R0(void);
	virtual void I_XCH_A_AT_R1(void);
	virtual void I_XCH_A_R0(void);
	virtual void I_XCH_A_R1(void);
	virtual void I_XCH_A_R2(void);
	virtual void I_XCH_A_R3(void);
	virtual void I_XCH_A_R4(void);
	virtual void I_XCH_A_R5(void);
	virtual void I_XCH_A_R6(void);
	virtual void I_XCH_A_R7(void);
	virtual void I_XCHD_A_AT_R0(void);
	virtual void I_XCHD_A_AT_R1(void);
	virtual void I_XRL_A_AT_R0(void);
	virtual void I_XRL_A_AT_R1(void);
	virtual void I_XRL_A_IMM(void);
	virtual void I_XRL_A_R0(void);
	virtual void I_XRL_A_R1(void);
	virtual void I_XRL_A_R2(void);
	virtual void I_XRL_A_R3(void);
	virtual void I_XRL_A_R4(void);
	virtual void I_XRL_A_R5(void);
	virtual void I_XRL_A_R6(void);
	virtual void I_XRL_A_R7(void);
	virtual void UpdatePSW(void);

public:
	//set instruction Mem Pointer
	virtual void SetInstrMemPtr(void* ptr);
	//set ROM length
	virtual void SetROMLen(unsigned long len);
	//set WD1772 instance
	virtual void SetWD1772(WD1772Base* myWD1772);
	//set Serial instance, this serial is bound to CPU and gets/puts data into global serial
	virtual void SetSerial(SerialBase* mySerial);
	//set instruction pointer
	virtual void SetInstrPtr(unsigned short myPtr);
	//get instruction Pointer
	virtual unsigned short GetInstrPtr();
	//resets the cpu
	virtual void ResetCPU();
	//causes cpu delay, is used by serial transfer
	virtual void DelayCPU(int64_t myNoOfCycles);
	//get current cpu cycle number
	virtual int64_t GetCPUCycle();
	//set drive number
	virtual void SetDriveNumber(unsigned char DriveNo);
protected:
	//set a value in instruction memory
	virtual void SetInstrMem(unsigned short addr, unsigned char value); //for testing only
	//get a value from instruction memory
	virtual unsigned char GetInstrMem(unsigned short addr);
	//for instructions changing port 1
	virtual void NotifyChangePort1(unsigned char oldPort, unsigned char newPort);

	friend class testCPU8040;

};
}
#endif
