// management of WD1772 data, base

#ifndef __XATREMU_WD1772_BASE__
#define __XATREMU_WD1772_BASE__

#ifdef WITH_CAPS
#include "xatremu_caps.h"
#endif
#include "testdefines.h"
#include "cpu8040_base.h"

namespace xatremu
{
class WD1772Base
{
public:
	WD1772Base();
	virtual ~WD1772Base() = 0;
	virtual void SetCPU(CPU8040Base* myCPU);
	virtual void ResetWD1772();
	virtual void ChooseDoubleDensity(bool doubleDensity);
	virtual void MasterReset();
	virtual void ChooseRegister(unsigned char reg); //address 0-3
	virtual void ChooseReadRegisters(bool readRegs); //read or write
	virtual void ChooseSide(bool side);
	virtual bool SetImage(char* imageName);
	virtual void EjectImage();
	virtual void Choose288RPM(bool rpm288);
	virtual bool GetIntRq();
	virtual bool StateT1Line(); //data request line
	virtual void CheckDRQEvent();
	virtual void CheckHeadStepDelay();
	virtual void CheckHeadCalmDelay();
	virtual void CheckMotor();
	virtual void MotorOn();
	virtual void MotorOff();
	virtual void PrepareMotorOff();
	virtual void CountTrackDataOffset();
	virtual unsigned char ByteFromWD1772();
	virtual void ByteToWD1772(unsigned char myByte);
	virtual unsigned char GetStatusRegister();
	virtual void CreateTrackData();
	virtual void CreateTrackFromATR();
	virtual unsigned int EncodeByteFM(unsigned char byteToEncode);
	virtual unsigned char DecodeByteFM(unsigned int byteToDecode);
	virtual unsigned int EncodeByteMFM(unsigned char byteToEncode, bool previousBit);
	virtual unsigned char DecodeByteMFM(unsigned int byteToDecode);
	virtual unsigned int CalculateCRC(unsigned int CRCStart, unsigned char byteToAdd);
	virtual unsigned int CalculateCRCEnd(unsigned int CRCStart);
	virtual void SetTrackDataWord(unsigned char * trackPointer, unsigned int dataWord);
	virtual unsigned long int GetTrackDataLongword(unsigned char * trackStartPointer, unsigned int dataOffset);
	virtual void DataSearchTimeOut();
	virtual unsigned int CyclesToNextDataWord();
	virtual signed char SearchSync(unsigned long int myDataLongWord, unsigned int mySync1, unsigned int mySync2);
	virtual void ResetTrackWriteProperties(bool noData);
	virtual void TransferTrackWriteProperties();
	virtual void LoadWATRData();
	virtual void FreeWATRData();
	virtual void SaveWATRData();
	virtual void SetWriteProtect(unsigned char WriteProtected);
	virtual unsigned char FloppyState();

#ifdef WITH_CAPS
	virtual void SetCapsImgId(SDWORD imgId);
	virtual char AnalyzeFromCT();
	virtual char AnalyzeFromIPF();
	virtual unsigned char DecodeByteFMCT(unsigned long byteToDecode);
	virtual unsigned int RecodeDataFMCT(unsigned long recodeLongWord);
	virtual uint64_t GetTrackDataQuadWordCT(char currentRevolution, unsigned int dataOffset);
	virtual unsigned long int GetTrackDataLongWordCT(char currentRevolution, unsigned int dataOffset);
	virtual signed char SearchSyncsCTFM(uint64_t myDataQuadWord, unsigned long mySync1, unsigned long mySync2, unsigned long mySync3);
#endif

protected:
	bool m_IsDoubleDensity;
	unsigned char m_SelectedRegister; // which address (0-3) is selected
	bool m_SelectRegsForRead; // true - read, false - write
	unsigned char m_LastCommand; //last command given
	bool m_IsDiskSide1; //false: side 0, true: side 1
	unsigned char m_WantedSector; //0-17 FM, 0-25 MFM/128, 0-17 MFM/256
	unsigned char m_WantedTrack; //0-39
	unsigned char m_CurrentTrack; //0-39
	unsigned int m_SectorLength; //128, 256
	bool m_StepIn; //true - step in, false - step out
	bool m_MotorIsOn;
	bool m_Busy; //busy flag
	bool m_RecordNotFound;
	bool m_CRCError;
	bool m_DeletedDataMark;
	bool m_IntRQ;
	bool m_DRQ;	//data request
	bool m_Is288RPM;	// for compatibility, create a WD1772 for 288RPM too
	bool m_Is40Track;	// for 40 track CT images

	//formula: (is288RPM ? 0.208333333 : 0.2) * 8333333.333 / bytesPerTrack * 2
	//unsigned short m_CPUCyclesToNextEncodedWordOfTrack;
	//better: Have total time since motor on tracked as double
	double m_totalTime;
	int64_t m_MotorStartCycle;
	CPU8040Base* m_CPU8040;
	unsigned char * m_ATRData; //data of a possible ATR image
	unsigned short int m_EncodedBytesPerTrack; //3255 for FM as standard
	unsigned short int m_currentTrackOffset; //current offset in track, important for changing tracks
	double m_OffsetStartTrack; //if track starts not at index, in seconds
	signed char m_SyncBitShift; //bit shift of sync in data stream
	unsigned short int m_CurrentCRC; //crc value in progress
	unsigned char m_headerData[7]; //header data as read from disk
	unsigned char m_CurrentByte; //current sector data byte as read or to write
	unsigned int m_WordsElapsed; //count of words in the current operation
	unsigned char * m_trackData; //data of an encoded track from an ATR image
	bool m_imageLoaded;
	unsigned char m_imageType; //0-NONE, 1-ATR, 2-RAW CT, 3-IPF
	unsigned char m_encodingType; //0-FM, 1-MFM-Medium, 2-MFM-Double-SingleSided, 3-MFM-DSDD for ATR
					// 0-FM, 1-MFM for CT Raw
	unsigned char* m_SectorDataOnWrite; //pointer to data write area
	unsigned char m_SectorWriteDummy[0x100];
	unsigned char m_CurrentSectorSequenceNumber;
	bool m_trackIsValid;
	bool m_DataByteProvided; //a byte has sent to the data register for fetching
	bool m_DataOverflow; //data overflow has occurred
	bool m_WriteProtect; //mechanical write protect of disk
	unsigned char m_FloppyState;

	struct WriteSectorProperties {
		unsigned char Dummy;
		unsigned char SectorSequenceNumber; //how often that sector number appears duplicate
		short int OffsetInProcessed; //offset of header of sector in trackdata
		unsigned char * SectorData; //data if different [i.e. written] from read
	};

	unsigned short int m_SectorLengthInTrack; //can be 0x80 or 0x100
	WriteSectorProperties m_SectorLayout[26][30]; // sector layout: [Sector number-1][occurrence]
	unsigned short int m_LastHeaderOffset;

#ifdef WITH_CAPS
	SDWORD m_CAPSImgId; //imageID for CAPS container
	UDWORD m_CAPSMinHead;  //min head in CT image
	UDWORD m_CAPSMaxHead;  //max head in CT image
	char m_CurrentRevolution; //revolution related to CT/ipf
	unsigned int m_OldRevolution; //revolutions since motor_on
	bool m_CAPSImgPresent;
	CapsTrackInfo m_CAPSTrackInfo[5];
	CapsTrackInfo * m_CAPSTrackInfoPtr[2];
#endif
	char m_GivenFileName[530];
	char m_WATRFileName[530];

	struct WATRSectors{
		unsigned char SectorNumber; // 0 - 25
		unsigned char SectorSequence; // 0 - ...
		unsigned short int SectorOffset; //0 - m_EncodedBytesPerTrack
		unsigned char * SectorData;
		unsigned long int FileOffsetSectorDescr;
		unsigned long int FileOffsetSectorData;
	};

	struct WATRTrack{
		WATRSectors * SectorPtr[30];
		unsigned long int FileOffset;
	};

	struct WATRSide{
		WATRTrack * TrackPtr[40];
		unsigned long int FileOffset;
	};

	struct WATRHeader{
		char W_ID[4];
		unsigned short int DiskSectorLen;
		WATRSide * SidePtr[2];
	};

	WATRHeader m_WATRHeader;



private:
	friend class testCPU8040;
};
}
#endif
